# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    Callable,
    Dict,
    List,
    Optional,
)
from numpy import (
    angle,
    array,
    ndarray,
)
import dearpygui.dearpygui as dpg
from deareis.signals import Signal
import deareis.signals as signals
from deareis.gui.plots import Bode, Nyquist
from deareis.utility import (
    align_numbers,
    format_number,
)
from deareis.tooltips import (
    attach_tooltip,
    update_tooltip,
)
import deareis.tooltips as tooltips
import deareis.themes as themes
from deareis.data import DataSet


class DataTable:
    def __init__(self):
        self._table: int = dpg.generate_uuid()
        with dpg.table(
            borders_outerV=True,
            borders_outerH=True,
            borders_innerV=True,
            borders_innerH=True,
            scrollY=True,
            freeze_rows=1,
            tag=self._table,
        ):
            dpg.add_table_column(
                label="Mask",
                width_fixed=True,
            )
            attach_tooltip(tooltips.data_sets.mask)
            dpg.add_table_column(
                label="Index",
                width_fixed=True,
            )
            dpg.add_table_column(
                label="f (Hz)",
            )
            attach_tooltip(tooltips.data_sets.frequency)
            dpg.add_table_column(
                label="Z' (ohm)",
            )
            attach_tooltip(tooltips.data_sets.real)
            dpg.add_table_column(
                label='-Z" (ohm)',
            )
            attach_tooltip(tooltips.data_sets.imaginary)
            dpg.add_table_column(
                label="|Z| (ohm)",
            )
            attach_tooltip(tooltips.data_sets.magnitude)
            dpg.add_table_column(
                label="-phi (°)",
            )
            attach_tooltip(tooltips.data_sets.phase)

    def clear(self):
        dpg.delete_item(self._table, children_only=True, slot=1)

    def populate(self, data: DataSet):
        assert type(data) is DataSet, data
        num_rows: int = data.get_num_points(masked=None)
        if num_rows == 0:
            self.clear()
            return
        rows: List[int] = dpg.get_item_children(self._table, slot=1)
        if len(rows) > num_rows:
            while len(rows) > num_rows:
                dpg.delete_item(rows.pop())
            return
        i: int
        for i in range(0, num_rows - len(rows)):
            with dpg.table_row(parent=self._table):
                dpg.add_checkbox(
                    default_value=False,
                    callback=lambda s, a, u: signals.emit(
                        Signal.TOGGLE_DATA_POINT,
                        state=a,
                        index=u[0],
                        data=u[1],
                    ),
                    user_data=(
                        i,
                        data,
                    ),
                )
                dpg.add_text("")
                dpg.add_text("")
                attach_tooltip("")
                dpg.add_text("")
                attach_tooltip("")
                dpg.add_text("")
                attach_tooltip("")
                dpg.add_text("")
                attach_tooltip("")
                dpg.add_text("")
                attach_tooltip("")

    def update(self, data: DataSet):
        assert type(data) is DataSet, data
        rows: List[int] = dpg.get_item_children(self._table, slot=1)
        assert len(rows) == data.get_num_points(masked=None)
        mask: Dict[int, bool] = data.get_mask()
        indices: List[str] = list(
            map(lambda _: str(_ + 1), range(0, data.get_num_points(masked=None)))
        )
        frequencies: ndarray = data.get_frequency(masked=None)
        freqs: List[str] = list(
            map(
                lambda _: format_number(_, significants=4),
                frequencies,
            )
        )
        Z: ndarray = data.get_impedance(masked=None)
        reals: List[str] = list(
            map(
                lambda _: format_number(
                    _.real,
                    significants=4,
                ),
                Z,
            )
        )
        imags: List[str] = list(
            map(lambda _: format_number(-_.imag, significants=4), Z)
        )
        mags: List[str] = list(
            map(
                lambda _: format_number(
                    abs(_),
                    significants=4,
                ),
                Z,
            )
        )
        phis: List[str] = list(
            map(
                lambda _: format_number(
                    -angle(_, deg=True),  # type: ignore
                    significants=4,
                    exponent=False,
                ),
                Z,
            )
        )
        indices = align_numbers(indices)
        freqs = align_numbers(freqs)
        reals = align_numbers(reals)
        imags = align_numbers(imags)
        mags = align_numbers(mags)
        phis = align_numbers(phis)
        fmt: str = "{:.6E}"
        i: int
        row: int
        for i, row in enumerate(rows):
            cells: List[int] = dpg.get_item_children(row, slot=1)
            # Mask
            dpg.configure_item(
                cells[0],
                default_value=mask.get(i, False),
                user_data=(
                    i,
                    data,
                ),
            )
            # Index
            dpg.configure_item(
                cells[1],
                default_value=indices[i],
            )
            # Frequency
            dpg.configure_item(
                cells[2],
                default_value=freqs[i],
            )
            update_tooltip(dpg.get_item_user_data(cells[3]), fmt.format(frequencies[i]))
            # Z, real
            dpg.configure_item(
                cells[4],
                default_value=reals[i],
            )
            update_tooltip(dpg.get_item_user_data(cells[5]), fmt.format(Z[i].real))
            # -Z, imaginary
            dpg.configure_item(
                cells[6],
                default_value=imags[i],
            )
            update_tooltip(dpg.get_item_user_data(cells[7]), fmt.format(-Z[i].imag))
            # Z, magnitude
            dpg.configure_item(
                cells[8],
                default_value=mags[i],
            )
            update_tooltip(dpg.get_item_user_data(cells[9]), fmt.format(abs(Z[i])))
            # Phase shift
            dpg.configure_item(
                cells[10],
                default_value=phis[i],
            )
            update_tooltip(
                dpg.get_item_user_data(cells[11]), fmt.format(-angle(Z[i], deg=True))
            )


class DataSetsTab:
    def __init__(self):
        self.queued_update: Optional[Callable] = None
        self.tab: int = dpg.generate_uuid()
        with dpg.tab(label="Data sets", tag=self.tab):
            with dpg.child_window(border=False, width=-1, height=-1):
                with dpg.group(horizontal=True):
                    self.table_window: int = dpg.generate_uuid()
                    self.table_width: int = 600
                    with dpg.child_window(
                        border=False,
                        width=self.table_width,
                        tag=self.table_window,
                        show=True,
                    ):
                        label_pad: int = 8
                        with dpg.child_window(border=True, height=82, width=-2):
                            with dpg.group(horizontal=True):
                                with dpg.child_window(border=False, width=-72):
                                    # TODO: Split into combo class?
                                    self.data_sets_combo: int = dpg.generate_uuid()
                                    with dpg.group(horizontal=True):
                                        self.visibility_item: int = dpg.generate_uuid()
                                        dpg.add_text(
                                            "Data set".rjust(label_pad),
                                            tag=self.visibility_item,
                                        )
                                        dpg.add_combo(
                                            callback=lambda s, a, u: signals.emit(
                                                Signal.SELECT_DATA_SET,
                                                data=u.get(a),
                                            ),
                                            user_data={},
                                            width=-1,
                                            tag=self.data_sets_combo,
                                        )
                                    self.label_input: int = dpg.generate_uuid()
                                    with dpg.group(horizontal=True):
                                        dpg.add_text("Label".rjust(label_pad))
                                        dpg.add_input_text(
                                            on_enter=True,
                                            callback=lambda s, a, u: signals.emit(
                                                Signal.RENAME_DATA_SET,
                                                label=a,
                                                data=u,
                                            ),
                                            width=-1,
                                            tag=self.label_input,
                                        )
                                    self.path_input: int = dpg.generate_uuid()
                                    with dpg.group(horizontal=True):
                                        dpg.add_text("Path".rjust(label_pad))
                                        dpg.add_input_text(
                                            on_enter=True,
                                            callback=lambda s, a, u: signals.emit(
                                                Signal.MODIFY_DATA_SET_PATH,
                                                path=a,
                                                data=u,
                                            ),
                                            width=-1,
                                            tag=self.path_input,
                                        )
                                with dpg.child_window(border=False, width=-1):
                                    self.load_button: int = dpg.generate_uuid()
                                    dpg.add_button(
                                        label="Load",
                                        callback=lambda s, a, u: signals.emit(
                                            Signal.SELECT_DATA_SET_FILES,
                                        ),
                                        width=-1,
                                        tag=self.load_button,
                                    )
                                    attach_tooltip(tooltips.data_sets.load)
                                    self.delete_button: int = dpg.generate_uuid()
                                    dpg.add_button(
                                        label="Delete",
                                        callback=lambda s, a, u: signals.emit(
                                            Signal.DELETE_DATA_SET,
                                            data=u,
                                        ),
                                        width=-1,
                                        tag=self.delete_button,
                                    )
                                    attach_tooltip(tooltips.data_sets.delete)
                                    self.average_button: int = dpg.generate_uuid()
                                    dpg.add_button(
                                        label="Average",
                                        callback=lambda s, a, u: signals.emit(
                                            Signal.SELECT_DATA_SETS_TO_AVERAGE,
                                        ),
                                        width=-1,
                                        tag=self.average_button,
                                    )
                                    attach_tooltip(tooltips.data_sets.average)
                        with dpg.child_window(
                            border=False, width=-2, height=-40, show=True
                        ):
                            self.data_table = DataTable()
                        with dpg.child_window(border=True, width=-2):
                            with dpg.group(horizontal=True):
                                self.toggle_points_button: int = dpg.generate_uuid()
                                dpg.add_button(
                                    label="Toggle points",
                                    tag=self.toggle_points_button,
                                    callback=lambda s, a, u: signals.emit(
                                        Signal.SELECT_DATA_POINTS_TO_TOGGLE,
                                        data=u,
                                    ),
                                )
                                attach_tooltip(tooltips.data_sets.toggle)
                                self.copy_mask_button: int = dpg.generate_uuid()
                                dpg.add_button(
                                    label="Copy mask",
                                    tag=self.copy_mask_button,
                                    callback=lambda s, a, u: signals.emit(
                                        Signal.SELECT_DATA_SET_MASK_TO_COPY,
                                        data=u,
                                    ),
                                )
                                attach_tooltip(tooltips.data_sets.copy)
                                self.subtract_impedance_button: int = (
                                    dpg.generate_uuid()
                                )
                                dpg.add_button(
                                    label="Subtract",
                                    tag=self.subtract_impedance_button,
                                    callback=lambda s, a, u: signals.emit(
                                        Signal.SELECT_IMPEDANCE_TO_SUBTRACT,
                                        data=u,
                                    ),
                                )
                                attach_tooltip(tooltips.data_sets.subtract)
                                self.enlarge_nyquist_button: int = dpg.generate_uuid()
                                self.adjust_nyquist_limits_checkbox: int = (
                                    dpg.generate_uuid()
                                )
                                dpg.add_button(
                                    label="Enlarge Nyquist",
                                    callback=self.show_enlarged_nyquist,
                                    tag=self.enlarge_nyquist_button,
                                )
                                dpg.add_checkbox(
                                    default_value=True,
                                    tag=self.adjust_nyquist_limits_checkbox,
                                )
                                attach_tooltip(tooltips.general.adjust_nyquist_limits)
                                self.enlarge_bode_button: int = dpg.generate_uuid()
                                self.adjust_bode_limits_checkbox: int = (
                                    dpg.generate_uuid()
                                )
                                dpg.add_button(
                                    label="Enlarge Bode",
                                    callback=self.show_enlarged_bode,
                                    tag=self.enlarge_bode_button,
                                )
                                dpg.add_checkbox(
                                    default_value=True,
                                    tag=self.adjust_bode_limits_checkbox,
                                )
                                attach_tooltip(tooltips.general.adjust_bode_limits)
                    self.plot_window: int = dpg.generate_uuid()
                    self.plot_width: int = 400
                    with dpg.child_window(
                        border=False,
                        width=-1,
                        no_scrollbar=True,
                        tag=self.plot_window,
                        show=True,
                    ):
                        self.nyquist_plot: Nyquist = Nyquist()
                        self.nyquist_plot.plot(
                            real=array([]),
                            imaginary=array([]),
                            label="Data",
                            line=False,
                            theme=themes.nyquist.data,
                        )
                        self.nyquist_plot.plot(
                            real=array([]),
                            imaginary=array([]),
                            label="Data",
                            line=True,
                            theme=themes.nyquist.data,
                            show_label=False,
                        )
                        self.bode_plot: Bode = Bode()
                        self.bode_plot.plot(
                            frequency=array([]),
                            magnitude=array([]),
                            phase=array([]),
                            labels=(
                                "|Z|",
                                "phi",
                            ),
                            line=False,
                            themes=(
                                themes.bode.magnitude_data,
                                themes.bode.phase_data,
                            ),
                        )
                        self.bode_plot.plot(
                            frequency=array([]),
                            magnitude=array([]),
                            phase=array([]),
                            labels=(
                                "|Z|",
                                "phi",
                            ),
                            line=True,
                            themes=(
                                themes.bode.magnitude_data,
                                themes.bode.phase_data,
                            ),
                            show_labels=False,
                        )

    def is_visible(self) -> bool:
        return dpg.is_item_visible(self.visibility_item)

    def resize(self, width: int, height: int):
        assert type(width) is int and width > 0, width
        assert type(height) is int and height > 0, height
        if not self.is_visible():
            return
        if width < (self.table_width + self.plot_width):
            if dpg.is_item_shown(self.plot_window):
                dpg.hide_item(self.plot_window)
                dpg.set_item_width(self.table_window, -1)
                dpg.set_item_label(self.enlarge_nyquist_button, "Show Nyquist")
                dpg.set_item_label(self.enlarge_bode_button, "Show Bode")
        else:
            if not dpg.is_item_shown(self.plot_window):
                dpg.show_item(self.plot_window)
                dpg.set_item_width(self.table_window, self.table_width)
                dpg.split_frame()
                dpg.set_item_label(self.enlarge_nyquist_button, "Enlarge Nyquist")
                dpg.set_item_label(self.enlarge_bode_button, "Enlarge Bode")
        if not dpg.is_item_shown(self.plot_window):
            return
        width, height = dpg.get_item_rect_size(self.plot_window)
        item: int
        for item in dpg.get_item_children(self.plot_window, slot=1):
            dpg.set_item_height(item, height / 2)

    def clear(self):
        dpg.set_value(self.data_sets_combo, "")
        dpg.set_value(self.label_input, "")
        dpg.set_value(self.path_input, "")
        dpg.set_item_user_data(self.delete_button, None)
        dpg.set_item_user_data(self.toggle_points_button, None)
        dpg.set_item_user_data(self.copy_mask_button, None)
        dpg.set_item_user_data(self.subtract_impedance_button, None)
        self.data_table.clear()
        self.nyquist_plot.clear(delete=False)
        self.bode_plot.clear(delete=False)

    def populate_data_sets(self, labels: List[str], lookup: Dict[str, DataSet]):
        assert type(labels) is list, labels
        assert type(lookup) is dict, lookup
        label: str = dpg.get_value(self.data_sets_combo) or ""
        if labels and label not in labels:
            label = labels[0]
        dpg.configure_item(
            self.data_sets_combo,
            items=labels,
            default_value=label,
            user_data=lookup,
        )

    def get_next_data_set(self) -> Optional[DataSet]:
        lookup: Dict[str, DataSet] = dpg.get_item_user_data(self.data_sets_combo)
        if not lookup:
            return None
        labels: List[str] = list(lookup.keys())
        index: int = labels.index(dpg.get_value(self.data_sets_combo)) + 1
        return lookup[labels[index % len(labels)]]

    def get_previous_data_set(self) -> Optional[DataSet]:
        lookup: Dict[str, DataSet] = dpg.get_item_user_data(self.data_sets_combo)
        if not lookup:
            return None
        labels: List[str] = list(lookup.keys())
        index: int = labels.index(dpg.get_value(self.data_sets_combo)) - 1
        return lookup[labels[index % len(labels)]]

    def select_data_set(self, data: Optional[DataSet]):
        assert type(data) is DataSet or data is None, data
        if not self.is_visible():
            self.queued_update = lambda: self.select_data_set(data)
            return
        self.queued_update = None
        if data is None:
            self.clear()
            return
        self.data_table.populate(data)
        self.data_table.update(data)
        dpg.set_value(self.data_sets_combo, data.get_label())
        dpg.configure_item(
            self.label_input,
            default_value=data.get_label(),
            user_data=data,
        )
        dpg.configure_item(
            self.path_input,
            default_value=data.get_path(),
            user_data=data,
        )
        # TODO: Could this be simplified so that data is stored in one place?
        dpg.set_item_user_data(self.delete_button, data)
        dpg.set_item_user_data(self.toggle_points_button, data)
        dpg.set_item_user_data(self.copy_mask_button, data)
        dpg.set_item_user_data(self.subtract_impedance_button, data)
        real: ndarray
        imag: ndarray
        real, imag = data.get_nyquist_data()
        self.nyquist_plot.update(
            index=0,
            real=real,
            imaginary=imag,
        )
        self.nyquist_plot.update(
            index=1,
            real=real,
            imaginary=imag,
        )
        freq: ndarray
        mag: ndarray
        phase: ndarray
        freq, mag, phase = data.get_bode_data()
        self.bode_plot.update(
            index=0,
            frequency=freq,
            magnitude=mag,
            phase=phase,
        )
        self.bode_plot.update(
            index=1,
            frequency=freq,
            magnitude=mag,
            phase=phase,
        )
        if dpg.get_value(self.adjust_nyquist_limits_checkbox):
            self.nyquist_plot.queue_limits_adjustment()
        if dpg.get_value(self.adjust_bode_limits_checkbox):
            self.bode_plot.queue_limits_adjustment()

    def show_enlarged_nyquist(self):
        signals.emit(
            Signal.SHOW_ENLARGED_PLOT,
            plot=self.nyquist_plot,
            adjust_limits=dpg.get_value(self.adjust_nyquist_limits_checkbox),
        )

    def show_enlarged_bode(self):
        signals.emit(
            Signal.SHOW_ENLARGED_PLOT,
            plot=self.bode_plot,
            adjust_limits=dpg.get_value(self.adjust_bode_limits_checkbox),
        )

    def has_active_input(self) -> bool:
        return dpg.is_item_active(self.label_input) or dpg.is_item_active(
            self.path_input
        )



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

import dearpygui.dearpygui as dpg
from deareis.signals import Signal
import deareis.signals as signals
from deareis.utility import calculate_window_position_dimensions
from .base import Plot
from .bode import (
    Bode,
    BodeMagnitude,
    BodePhase,
)
from .image import Image
from .mu_xps import MuXps
from .nyquist import Nyquist
from .residuals import Residuals
from .drt import DRT
from .impedance import (
    Impedance,
    ImpedanceReal,
    ImpedanceImaginary,
)


class ModalPlotWindow:
    def __init__(self, original: Plot, adjust_limits: bool):
        labels: dict = {
            Bode: "Bode",
            BodeMagnitude: "Bode - magnitude",
            BodePhase: "Bode - phase",
            Nyquist: "Nyquist",
            Residuals: "Residuals",
            DRT: "Distribution of relaxation times",
            Impedance: "Complex impedance",
            ImpedanceImaginary: "Imaginary impedance",
            ImpedanceReal: "Real impedance",
        }
        x: int
        y: int
        w: int
        h: int
        x, y, w, h = calculate_window_position_dimensions()
        self.window: int = dpg.generate_uuid()
        self.key_handler = dpg.generate_uuid()
        with dpg.window(
            label=labels.get(type(original), "Unknown plot type"),
            modal=True,
            pos=(
                x,
                y,
            ),
            width=w,
            height=h,
            on_close=self.close,
            tag=self.window,
        ):
            copy: Plot = type(original).duplicate(original)
            if adjust_limits:
                copy.queue_limits_adjustment()
            else:
                copy.copy_limits(original)
        with dpg.handler_registry(tag=self.key_handler):
            dpg.add_key_release_handler(key=dpg.mvKey_Escape, callback=self.close)

    def close(self):
        dpg.delete_item(self.key_handler)
        dpg.delete_item(self.window)
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)


def show_modal_plot_window(original: Plot, adjust_limits: bool = True):
    assert hasattr(original, "duplicate"), type(original)
    assert type(adjust_limits) is bool, adjust_limits
    modal_window: ModalPlotWindow = ModalPlotWindow(original, adjust_limits)
    signals.emit(
        Signal.BLOCK_KEYBINDINGS,
        window=modal_window.window,
        window_object=modal_window,
    )



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from random import choice
from types import SimpleNamespace
from typing import Dict, List, Set, Tuple
import dearpygui.dearpygui as dpg


PLOT_MARKERS: Dict[str, int] = {
    "Arrow down": dpg.mvPlotMarker_Down,
    "Arrow left": dpg.mvPlotMarker_Left,
    "Arrow right": dpg.mvPlotMarker_Right,
    "Arrow up": dpg.mvPlotMarker_Up,
    "Asterisk": dpg.mvPlotMarker_Asterisk,
    "Circle": dpg.mvPlotMarker_Circle,
    "Cross": dpg.mvPlotMarker_Cross,
    "Diamond": dpg.mvPlotMarker_Diamond,
    "Plus": dpg.mvPlotMarker_Plus,
    "Square": dpg.mvPlotMarker_Square,
}


VIBRANT_COLORS: List[List[float]] = [
    [  # Blue
        0.0,
        119.0,
        187.0,
        255.0,
    ],
    [  # Teal
        0.0,
        153.0,
        136.0,
        255.0,
    ],
    [  # Orange
        238.0,
        119.0,
        51.0,
        255.0,
    ],
    [  # Magenta
        238.0,
        51.0,
        119.0,
        255.0,
    ],
    [  # Red
        204.0,
        51.0,
        17.0,
        255.0,
    ],
    [  # Gray
        187.0,
        187.0,
        187.0,
        255.0,
    ],
    [  # Cyan
        51.0,
        187.0,
        238.0,
        255.0,
    ],
]


def get_plot_series_theme_color(parent: int) -> List[float]:
    assert type(parent) is int, parent
    color: List[float] = []
    item: int
    for item in dpg.get_item_children(parent, slot=1):
        i: int
        for i in dpg.get_item_children(item, slot=1):
            if not dpg.get_item_type(i).endswith("::mvThemeColor"):
                continue
            color = dpg.get_value(i)
    assert (
        type(color) is list and len(color) == 4 and all(map(lambda _: _ >= 0.0, color))
    ), color
    return color


def get_plot_series_theme_marker(parent: int) -> int:
    assert type(parent) is int, parent
    marker: int = -1
    item: int
    for item in dpg.get_item_children(parent, slot=1):
        i: int
        for i in dpg.get_item_children(item, slot=1):
            if not dpg.get_item_type(i).endswith("::mvThemeStyle"):
                continue
            marker = int(dpg.get_value(i)[0])
    assert type(marker) is int and (
        marker in PLOT_MARKERS.values() or marker == -1
    ), marker
    return marker


def get_random_color_marker(existing_themes: Dict[str, int]) -> Tuple[List[float], int]:
    available_colors: List[List[float]] = VIBRANT_COLORS[:]
    available_markers: List[int] = list(PLOT_MARKERS.values())
    existing_colors: List[List[float]] = []
    existing_markers: List[int] = []
    existing_combinations: List[str] = []
    color: List[float]
    marker: int
    uuid: str
    item: int
    for uuid, item in existing_themes.items():
        color = get_plot_series_theme_color(item)
        marker = get_plot_series_theme_marker(item)
        existing_colors.append(color)
        existing_markers.append(marker)
        existing_combinations.append(",".join(map(str, color)) + str(marker))
    if len(available_markers) > len(existing_markers):
        available_markers = list(set(available_markers) - set(existing_markers))
        ac: Set[str] = set(map(lambda _: ",".join(map(str, _)), available_colors))
        ec: Set[str] = set(map(lambda _: ",".join(map(str, _)), existing_colors))
        ac = ac - ec
        if len(ac) > 0:
            available_colors = list(map(lambda _: list(map(float, _.split(","))), ac))
            return (
                choice(available_colors),
                choice(available_markers),
            )
    possible_combinations: Dict[str, Tuple[List[float], int]] = {}
    combination: str
    for marker in PLOT_MARKERS.values():
        for color in VIBRANT_COLORS:
            combination = ",".join(map(str, color)) + str(marker)
            if combination not in existing_combinations:
                possible_combinations[combination] = (
                    color,
                    marker,
                )
    if len(possible_combinations) > 0:
        return choice(list(possible_combinations.values()))
    return (
        [255.0, 255.0, 255.0, 255.0],
        dpg.mvPlotMarker_Circle,
    )


def create_plot_series_theme(
    color: List[float],
    marker: int,
    tag: int = -1,
) -> int:
    assert (
        type(color) is list
        and len(color) == 4
        and all(map(lambda _: type(_) is float, color))
    ), color
    assert type(marker) is int and marker in PLOT_MARKERS.values(), marker
    assert type(tag) is int
    if tag < 0:
        tag = dpg.generate_uuid()
    R: float
    G: float
    B: float
    A: float
    R, G, B, A = color
    with dpg.theme(tag=tag):
        with dpg.theme_component(dpg.mvScatterSeries):
            dpg.add_theme_color(
                dpg.mvPlotCol_MarkerFill,
                (
                    R,
                    G,
                    B,
                    0.0,
                ),
                category=dpg.mvThemeCat_Plots,
            )
            dpg.add_theme_color(
                dpg.mvPlotCol_Line,
                color,
                category=dpg.mvThemeCat_Plots,
            )
            dpg.add_theme_color(
                dpg.mvPlotCol_MarkerOutline,
                color,
                category=dpg.mvThemeCat_Plots,
            )
            dpg.add_theme_style(
                dpg.mvPlotStyleVar_Marker, marker, category=dpg.mvThemeCat_Plots
            )
        with dpg.theme_component(dpg.mvLineSeries):
            dpg.add_theme_color(
                dpg.mvPlotCol_Line,
                color,
                category=dpg.mvThemeCat_Plots,
            )
        with dpg.theme_component(dpg.mvShadeSeries):
            dpg.add_theme_color(
                dpg.mvPlotCol_Fill,
                color,
                category=dpg.mvThemeCat_Plots,
            )
    return tag


def update_plot_series_theme_color(parent: int, color: List[float]):
    assert type(parent) is int, parent
    if type(color) is tuple:
        color = list(color)
    assert (
        type(color) is list
        and len(color) == 4
        and all(map(lambda _: type(_) is float, color))
    ), color
    uuids: List[int] = []
    item: int
    for item in dpg.get_item_children(parent, slot=1):
        i: int
        for i in dpg.get_item_children(item, slot=1):
            if not dpg.get_item_type(i).endswith("::mvThemeColor"):
                continue
            uuids.append(i)
    if len(uuids) == 0:
        return
    alpha: float = color[-1]
    color[-1] = 0
    dpg.set_value(uuids.pop(0), color)
    color[-1] = alpha
    while uuids:
        dpg.set_value(uuids.pop(0), color)


def update_plot_series_theme_marker(parent: int, marker: int):
    assert type(parent) is int, parent
    assert type(marker) is int, marker
    item: int
    for item in dpg.get_item_children(parent, slot=1):
        i: int
        for i in dpg.get_item_children(item, slot=1):
            if not dpg.get_item_type(i).endswith("::mvThemeStyle"):
                continue
            dpg.set_value(i, [marker, -1])


nyquist: SimpleNamespace = SimpleNamespace(
    **{
        "data": create_plot_series_theme(
            [51.0, 187.0, 238.0, 190.0],
            dpg.mvPlotMarker_Circle,
        ),
        "simulation": create_plot_series_theme(
            [238.0, 51.0, 119.0, 190.0],
            dpg.mvPlotMarker_Cross,
        ),
    }
)

bode: SimpleNamespace = SimpleNamespace(
    **{
        "magnitude_data": create_plot_series_theme(
            [51.0, 187.0, 238.0, 190.0],
            dpg.mvPlotMarker_Circle,
        ),
        "magnitude_simulation": create_plot_series_theme(
            [238.0, 51.0, 119.0, 190.0],
            dpg.mvPlotMarker_Cross,
        ),
        "phase_data": create_plot_series_theme(
            [238.0, 119.0, 51.0, 190.0],
            dpg.mvPlotMarker_Square,
        ),
        "phase_simulation": create_plot_series_theme(
            [0.0, 153.0, 136.0, 190.0],
            dpg.mvPlotMarker_Plus,
        ),
    }
)


drt: SimpleNamespace = SimpleNamespace(
    **{
        "real_gamma": create_plot_series_theme(
            [51.0, 187.0, 238.0, 190.0],
            dpg.mvPlotMarker_Circle,
        ),
        "mean_gamma": create_plot_series_theme(
            [238.0, 119.0, 51.0, 190.0],
            dpg.mvPlotMarker_Circle,
        ),
        "credible_intervals": create_plot_series_theme(
            [238.0, 119.0, 51.0, 48.0],
            dpg.mvPlotMarker_Circle,
        ),
        "imaginary_gamma": create_plot_series_theme(
            [238.0, 119.0, 51.0, 190.0],
            dpg.mvPlotMarker_Circle,
        ),
    }
)


impedance: SimpleNamespace = SimpleNamespace(
    **{
        "real_data": create_plot_series_theme(
            [51.0, 187.0, 238.0, 190.0],
            dpg.mvPlotMarker_Circle,
        ),
        "real_simulation": create_plot_series_theme(
            [238.0, 51.0, 119.0, 190.0],
            dpg.mvPlotMarker_Cross,
        ),
        "imaginary_data": create_plot_series_theme(
            [238.0, 119.0, 51.0, 190.0],
            dpg.mvPlotMarker_Square,
        ),
        "imaginary_simulation": create_plot_series_theme(
            [0.0, 153.0, 136.0, 190.0],
            dpg.mvPlotMarker_Plus,
        ),
    }
)

residuals: SimpleNamespace = SimpleNamespace(
    **{
        "real": create_plot_series_theme(
            [238.0, 51.0, 119.0, 190.0],
            dpg.mvPlotMarker_Circle,
        ),
        "imaginary": create_plot_series_theme(
            [0.0, 153.0, 136.0, 190.0],
            dpg.mvPlotMarker_Square,
        ),
    }
)

mu_Xps: SimpleNamespace = SimpleNamespace(
    **{
        "mu": create_plot_series_theme(
            [238.0, 51.0, 119.0, 190.0],
            dpg.mvPlotMarker_Circle,
        ),
        "mu_highlight": create_plot_series_theme(
            [51.0, 187.0, 238.0, 190.0],
            dpg.mvPlotMarker_Circle,
        ),
        "Xps": create_plot_series_theme(
            [0.0, 153.0, 136.0, 190.0],
            dpg.mvPlotMarker_Square,
        ),
        "Xps_highlight": create_plot_series_theme(
            [238.0, 119.0, 51.0, 190.0],
            dpg.mvPlotMarker_Square,
        ),
        "mu_criterion": create_plot_series_theme(
            [255.0, 255.0, 255.0, 128.0],
            dpg.mvPlotMarker_Circle,
        ),
    }
)

plot: int
with dpg.theme() as plot:
    with dpg.theme_component(dpg.mvAll):
        dpg.add_theme_color(
            dpg.mvPlotCol_FrameBg,
            [
                255.0,
                255.0,
                255.0,
                0.0,
            ],
            category=dpg.mvThemeCat_Plots,
        )


_clean_tab: int = dpg.generate_uuid()
_dirty_tab: int = dpg.generate_uuid()
with dpg.theme(tag=_clean_tab):
    with dpg.theme_component(dpg.mvTab):
        dpg.add_theme_color(
            dpg.mvThemeCol_TabActive,
            [
                0.0,
                119.0,
                200.0,
                153.0,
            ],
        )
        dpg.add_theme_color(
            dpg.mvThemeCol_TabHovered,
            [
                29.0,
                151.0,
                236.0,
                103.0,
            ],
        )
        dpg.add_theme_color(
            dpg.mvThemeCol_Tab,
            [
                51.0,
                51.0,
                55.0,
                255.0,
            ],
        )
        dpg.add_theme_color(
            dpg.mvThemeCol_TabUnfocused,
            [
                51.0,
                51.0,
                55.0,
                255.0,
            ],
        )
        dpg.add_theme_color(
            dpg.mvThemeCol_TabUnfocusedActive,
            [
                0.0,
                119.0,
                200.0,
                153.0,
            ],
        )
with dpg.theme(tag=_dirty_tab):
    with dpg.theme_component(dpg.mvTab):
        dpg.add_theme_color(
            dpg.mvThemeCol_TabActive,
            [
                236.0,
                32.0,
                29.0,
                103.0,
            ],
        )
        dpg.add_theme_color(
            dpg.mvThemeCol_TabHovered,
            [
                236.0,
                80.0,
                80.0,
                153.0,
            ],
        )
        dpg.add_theme_color(
            dpg.mvThemeCol_Tab,
            [
                138.0,
                61.0,
                61.0,
                100.0,
            ],
        )
        dpg.add_theme_color(
            dpg.mvThemeCol_TabUnfocused,
            [
                138.0,
                61.0,
                61.0,
                100.0,
            ],
        )
        dpg.add_theme_color(
            dpg.mvThemeCol_TabUnfocusedActive,
            [
                236.0,
                32.0,
                29.0,
                103.0,
            ],
        )
tab: SimpleNamespace = SimpleNamespace(
    **{
        "clean": _clean_tab,
        "dirty": _dirty_tab,
    }
)


_valid_cdc: int = dpg.generate_uuid()
_invalid_cdc: int = dpg.generate_uuid()
with dpg.theme(tag=_valid_cdc):
    with dpg.theme_component(dpg.mvInputText, enabled_state=True):
        dpg.add_theme_color(
            dpg.mvThemeCol_Text,
            [
                255.0,
                255.0,
                255.0,
                255.0,
            ],
        )
    with dpg.theme_component(dpg.mvInputText, enabled_state=False):
        dpg.add_theme_color(
            dpg.mvThemeCol_Text,
            [
                255.0,
                255.0,
                255.0,
                255.0,
            ],
        )
with dpg.theme(tag=_invalid_cdc):
    with dpg.theme_component(dpg.mvInputText, enabled_state=True):
        dpg.add_theme_color(
            dpg.mvThemeCol_Text,
            [
                255.0,
                0.0,
                0.0,
                255.0,
            ],
        )
    with dpg.theme_component(dpg.mvInputText, enabled_state=False):
        dpg.add_theme_color(
            dpg.mvThemeCol_Text,
            [
                255.0,
                0.0,
                0.0,
                255.0,
            ],
        )
cdc: SimpleNamespace = SimpleNamespace(
    **{
        "valid": _valid_cdc,
        "invalid": _invalid_cdc,
    }
)


_valid_result: int = dpg.generate_uuid()
_invalid_result: int = dpg.generate_uuid()
with dpg.theme(tag=_valid_result):
    with dpg.theme_component(dpg.mvText, enabled_state=True):
        dpg.add_theme_color(
            dpg.mvThemeCol_Text,
            [
                255.0,
                255.0,
                255.0,
                255.0,
            ],
        )
    with dpg.theme_component(dpg.mvText, enabled_state=False):
        dpg.add_theme_color(
            dpg.mvThemeCol_Text,
            [
                255.0,
                255.0,
                255.0,
                255.0,
            ],
        )
with dpg.theme(tag=_invalid_result):
    with dpg.theme_component(dpg.mvText, enabled_state=True):
        dpg.add_theme_color(
            dpg.mvThemeCol_Text,
            [
                255.0,
                0.0,
                0.0,
                255.0,
            ],
        )
    with dpg.theme_component(dpg.mvText, enabled_state=False):
        dpg.add_theme_color(
            dpg.mvThemeCol_Text,
            [
                255.0,
                0.0,
                0.0,
                255.0,
            ],
        )
result: SimpleNamespace = SimpleNamespace(
    **{
        "valid": _valid_result,
        "invalid": _invalid_result,
    }
)


_valid_node: int = dpg.generate_uuid()
with dpg.theme(tag=_valid_node):
    with dpg.theme_component(dpg.mvNode):
        dpg.add_theme_color(
            dpg.mvNodeCol_NodeBackground,
            (
                255.0,
                255.0,
                255.0,
                255.0,
            ),
        )
_invalid_node: int = dpg.generate_uuid()
with dpg.theme(tag=_invalid_node):
    with dpg.theme_component(dpg.mvNode):
        dpg.add_theme_color(
            dpg.mvNodeCol_NodeBackground,
            (
                255,
                0.0,
                0.0,
                255.0,
            ),
        )
circuit_editor: SimpleNamespace = SimpleNamespace(
    **{
        "valid_node": _valid_node,
        "invalid_node": _invalid_node,
    }
)


_command_palette_result_highlighted: int
with dpg.theme() as _command_palette_result_highlighted:
    with dpg.theme_component(dpg.mvAll):
        dpg.add_theme_color(
            dpg.mvThemeCol_Text,
            [255.0, 255.0, 255.0, 255.0],
            category=dpg.mvThemeCat_Core,
        )
        dpg.add_theme_color(
            dpg.mvThemeCol_Button,
            [0.0, 119.0, 200.0, 100.0],
            category=dpg.mvThemeCat_Core,
        )
        dpg.add_theme_style(
            dpg.mvStyleVar_ButtonTextAlign,
            0.0,
            0.5,
            category=dpg.mvThemeCat_Core,
        )

_command_palette_result: int
with dpg.theme() as _command_palette_result:
    with dpg.theme_component(dpg.mvAll):
        dpg.add_theme_color(
            dpg.mvThemeCol_Text,
            [255.0, 255.0, 255.0, 255.0],
            category=dpg.mvThemeCat_Core,
        )
        dpg.add_theme_color(
            dpg.mvThemeCol_Button,
            [51.0, 51.0, 55.0, 255.0],
            category=dpg.mvThemeCat_Core,
        )
        dpg.add_theme_style(
            dpg.mvStyleVar_ButtonTextAlign,
            0.0,
            0.5,
            category=dpg.mvThemeCat_Core,
        )


command_palette: SimpleNamespace = SimpleNamespace(
    **{
        "result": _command_palette_result,
        "result_highlighted": _command_palette_result_highlighted,
    }
)


_file_dialog_folder: int
with dpg.theme() as _file_dialog_folder:
    with dpg.theme_component(dpg.mvAll):
        dpg.add_theme_style(
            dpg.mvStyleVar_ButtonTextAlign,
            0.0,
            0.5,
            category=dpg.mvThemeCat_Core,
        )


_file_dialog_file: int
with dpg.theme() as _file_dialog_file:
    with dpg.theme_component(dpg.mvAll):
        dpg.add_theme_style(
            dpg.mvStyleVar_ButtonTextAlign,
            0.0,
            0.5,
            category=dpg.mvThemeCat_Core,
        )


file_dialog: SimpleNamespace = SimpleNamespace(
    **{
        "folder_button": _file_dialog_folder,
        "file_button": _file_dialog_file,
    }
)


transparent_modal_background: int
with dpg.theme() as transparent_modal_background:
    with dpg.theme_component(dpg.mvAll):
        dpg.add_theme_color(
            dpg.mvThemeCol_ModalWindowDimBg,
            [0.0, 0.0, 0.0, 0.0],
            category=dpg.mvThemeCat_Core,
        )


# URL theme
url_theme: int
with dpg.theme() as url_theme:
    with dpg.theme_component(dpg.mvAll, enabled_state=True):
        dpg.add_theme_style(
            dpg.mvStyleVar_ButtonTextAlign,
            0.0,
            0.5,
            category=dpg.mvThemeCat_Core,
        )


# Global theme
_greyed_out: List[float] = [
    100.0,
    100.0,
    100.0,
    255.0,
]
_rounding: int = 2
global_theme: int
with dpg.theme() as global_theme:
    # Enabled
    with dpg.theme_component(dpg.mvAll):
        dpg.add_theme_style(
            dpg.mvStyleVar_WindowRounding, _rounding, category=dpg.mvThemeCat_Core
        )
        dpg.add_theme_style(
            dpg.mvStyleVar_ChildRounding, _rounding, category=dpg.mvThemeCat_Core
        )
        dpg.add_theme_style(
            dpg.mvStyleVar_FrameRounding, _rounding, category=dpg.mvThemeCat_Core
        )
        dpg.add_theme_style(
            dpg.mvStyleVar_PopupRounding, _rounding, category=dpg.mvThemeCat_Core
        )
        dpg.add_theme_style(
            dpg.mvStyleVar_ScrollbarRounding, _rounding, category=dpg.mvThemeCat_Core
        )
        dpg.add_theme_style(
            dpg.mvStyleVar_GrabRounding, _rounding, category=dpg.mvThemeCat_Core
        )
        dpg.add_theme_style(
            dpg.mvStyleVar_TabRounding, _rounding, category=dpg.mvThemeCat_Core
        )

        # Plots
        dpg.add_theme_style(
            dpg.mvPlotStyleVar_PlotPadding, 8, 8, category=dpg.mvThemeCat_Plots
        )
        dpg.add_theme_style(
            dpg.mvPlotStyleVar_LabelPadding, 5, 2, category=dpg.mvThemeCat_Plots
        )
        dpg.add_theme_style(
            dpg.mvPlotStyleVar_LegendPadding, 5, 5, category=dpg.mvThemeCat_Plots
        )
        dpg.add_theme_style(
            dpg.mvPlotStyleVar_LegendInnerPadding,
            4,
            4,
            category=dpg.mvThemeCat_Plots,
        )
        dpg.add_theme_style(
            dpg.mvPlotStyleVar_MousePosPadding, 8, 8, category=dpg.mvThemeCat_Plots
        )
        dpg.add_theme_style(
            dpg.mvPlotStyleVar_LineWeight, 1.5, category=dpg.mvThemeCat_Plots
        )
        dpg.add_theme_style(
            dpg.mvPlotStyleVar_MarkerSize, 4.5, category=dpg.mvThemeCat_Plots
        )
        dpg.add_theme_style(
            dpg.mvPlotStyleVar_MarkerWeight, 2.0, category=dpg.mvThemeCat_Plots
        )

        # Nodes
        dpg.add_theme_color(
            dpg.mvNodeCol_GridLine,
            [255.0, 255.0, 255.0, 15.0],
            category=dpg.mvThemeCat_Nodes,
        )

    # Disabled
    with dpg.theme_component(dpg.mvText, enabled_state=False):
        dpg.add_theme_color(dpg.mvThemeCol_Text, _greyed_out)

    with dpg.theme_component(dpg.mvSliderFloat, enabled_state=False):
        dpg.add_theme_color(dpg.mvThemeCol_Text, _greyed_out)
        dpg.add_theme_color(dpg.mvThemeCol_SliderGrab, _greyed_out)

    with dpg.theme_component(dpg.mvCombo, enabled_state=False):
        dpg.add_theme_color(dpg.mvThemeCol_Text, _greyed_out)

    with dpg.theme_component(dpg.mvCheckbox, enabled_state=False):
        dpg.add_theme_color(dpg.mvThemeCol_Text, _greyed_out)
        dpg.add_theme_color(dpg.mvThemeCol_CheckMark, _greyed_out)

    with dpg.theme_component(dpg.mvInputFloat, enabled_state=False):
        dpg.add_theme_color(dpg.mvThemeCol_Text, _greyed_out)
        dpg.add_theme_color(dpg.mvThemeCol_Button, _greyed_out)

    with dpg.theme_component(dpg.mvInputInt, enabled_state=False):
        dpg.add_theme_color(dpg.mvThemeCol_Text, _greyed_out)
        dpg.add_theme_color(dpg.mvThemeCol_Button, _greyed_out)

    with dpg.theme_component(dpg.mvInputText, enabled_state=False):
        dpg.add_theme_color(dpg.mvThemeCol_Text, _greyed_out)
        dpg.add_theme_color(dpg.mvThemeCol_Button, _greyed_out)

    with dpg.theme_component(dpg.mvButton, enabled_state=False):
        dpg.add_theme_color(dpg.mvThemeCol_Text, _greyed_out)

dpg.bind_theme(global_theme)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import Dict, List, Set, Tuple, Optional
import dearpygui.dearpygui as dpg
from numpy import array, ndarray
from deareis.signals import Signal
import deareis.signals as signals
from deareis.enums import Action, Context, action_contexts, action_descriptions
from deareis.utility import calculate_window_position_dimensions
from deareis.data.project import Project
from deareis.gui.project import ProjectTab
from deareis.keybindings import KeybindingHandler
import deareis.themes as themes
from deareis.tooltips import attach_tooltip
import deareis.tooltips as tooltips


class CommandPalette:
    def __init__(self, keybinding_handler: KeybindingHandler):
        assert type(keybinding_handler), keybinding_handler
        self.keybinding_handler: KeybindingHandler = keybinding_handler
        self.action_contexts: Dict[Action, Set[Context]] = {
            k: set(v)
            for k, v in action_contexts.items()
            if k != Action.SHOW_COMMAND_PALETTE
        }
        self.valid_actions: List[Tuple[Action, str]] = []
        self.action_history: List[Action] = []
        self.window: int = dpg.generate_uuid()
        with dpg.window(
            label="Command palette",
            no_close=True,
            modal=True,
            no_resize=True,
            show=False,
            tag=self.window,
        ):
            self.filter_input: int = dpg.generate_uuid()
            dpg.add_input_text(
                hint="Search...",
                tag=self.filter_input,
                width=-1,
                callback=lambda s, a, u: self.filter_results(a.strip().lower()),
            )
            attach_tooltip(tooltips.general.command_palette)
            self.results_table: int = dpg.generate_uuid()
            self.num_rows: int = 9
            with dpg.table(
                borders_outerV=True,
                borders_outerH=True,
                borders_innerV=True,
                borders_innerH=True,
                header_row=False,
                height=self.num_rows * 23,
                tag=self.results_table,
            ):
                dpg.add_table_column()
                for i in range(0, self.num_rows):
                    with dpg.table_row():
                        dpg.add_button(
                            label="",
                            callback=lambda s, a, u: self.perform_action(u, True),
                            width=-1,
                        )

    def show(self, contexts: List[Context], project: Project, tab: ProjectTab):
        self.current_contexts = contexts
        self.current_project: Project = project
        self.current_tab: ProjectTab = tab
        self.valid_actions = []
        contexts_set: Set[Context] = set(contexts)
        action: Action
        cons: Set[Context]
        for action, cons in self.action_contexts.items():
            if len(cons.intersection(contexts_set)) == 0:
                continue
            description: str = action_descriptions[action].split("\n")[0]
            if description.endswith(":"):
                description = description[:-1]
            elif description.endswith("."):
                description = description[:-1]
            self.valid_actions.append(
                (
                    action,
                    description,
                )
            )
        signals.emit(Signal.BLOCK_KEYBINDINGS, window=self.window, window_object=self)
        dpg.show_item(self.window)
        dpg.split_frame()
        x: int
        y: int
        w: int
        h: int = 24 + self.num_rows * 23 + 34
        x, y, w, h = calculate_window_position_dimensions(720, h)
        dpg.configure_item(
            self.window,
            pos=(
                x,
                y,
            ),
            width=w,
            height=h,
        )
        dpg.set_value(self.filter_input, "")
        self.filter_results("")
        self.key_handler: int = dpg.generate_uuid()
        with dpg.handler_registry(tag=self.key_handler):
            dpg.add_key_release_handler(
                key=dpg.mvKey_Escape,
                callback=self.hide,
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Return,
                callback=lambda s, a, u: self.perform_action(),
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Up,
                callback=lambda: self.navigate_result(step=-1),
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Prior,
                callback=lambda: self.navigate_result(step=-5),
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Home,
                callback=lambda: self.navigate_result(index=0),
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Down,
                callback=lambda: self.navigate_result(step=1),
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Next,
                callback=lambda: self.navigate_result(step=5),
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_End,
                callback=lambda: self.navigate_result(
                    index=len(self.valid_actions) - 1
                ),
            )
        dpg.split_frame()
        dpg.focus_item(self.filter_input)

    def hide(self):
        dpg.delete_item(self.key_handler)
        dpg.hide_item(self.window)
        dpg.split_frame()
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)

    def get_consecutive_letter_indices(self, source: str, target: str) -> List[int]:
        indices: List[int] = []
        letter: str
        for letter in source:
            index: int = -1
            if not indices or indices[-1] >= 0:
                index = target.find(letter, indices[-1] if indices else 0)
            indices.append(index)
        return indices

    def get_distances_from_word_beginning(
        self, letter_indices: List[int], target: str
    ) -> Tuple[List[int], List[int]]:
        if not letter_indices:
            return (
                [],
                [],
            )
        words: List[str] = target.split(" ")
        if not words:
            return (
                [],
                [],
            )
        distances: List[int] = []
        word_lengths: List[int] = []
        letter_index: int
        for letter_index in letter_indices:
            space_index: int = letter_index - 1
            letter: str = target[space_index]
            while letter != " " and space_index > 0:
                space_index -= 1
                letter = target[space_index]
            space_index += 1
            distances.append(letter_index - space_index)
            end_index: int = target.find(" ", space_index)
            if end_index < 0:
                end_index = len(target)
            word_lengths.append(end_index - space_index)
        return (
            distances,
            word_lengths,
        )

    def score_result(self, source: str, target: str, action: Action) -> float:
        score: float = 0.0
        num_valid_actions: int = len(self.valid_actions)
        if source == "":
            score -= int(action) + 1
            if action in self.action_history:
                recency_bonus: float = float(num_valid_actions)
                score += (
                    len(self.action_history) - self.action_history.index(action)
                ) * recency_bonus
        else:
            if source in target:
                score += 100
            len_source: int = len(source)
            len_target: int = len(target)
            indices: List[int] = self.get_consecutive_letter_indices(source, target)
            num_matches: int = len(list(filter(lambda _: _ >= 0, indices)))
            # Penalize for source letters not found in the target
            score -= abs(sum(filter(lambda _: _ < 0, indices))) * 25.0
            # Reward for source letters found in the target based on proximity to the beginning
            score += (
                sum(
                    map(
                        lambda _: (len_target - _) / len_target,
                        filter(lambda _: _ >= 0, indices),
                    )
                )
                * 100.0
            )
            # Penalize the difference between the number of found source letters and the length
            # of the target and normalize using the length of the source
            score -= (len_target - num_matches) / len_source * 10.0
            # Reward for the proximity of found source letters to the beginning of the word in
            # which the letter was found
            distances: List[int]
            word_lengths: List[int]
            distances, word_lengths = self.get_distances_from_word_beginning(
                list(filter(lambda _: _ >= 0, indices)), target
            )
            for distance, word_length in zip(distances, word_lengths):
                score += (word_length - distance) / word_length * 15.0
        return score

    def filter_results(self, string: str):
        self.result_index = 0
        scores: Dict[Action, float] = {}
        for (action, description) in self.valid_actions:
            scores[action] = self.score_result(
                string,
                description.lower(),
                action,
            )
        self.valid_actions.sort(key=lambda _: scores[_[0]], reverse=True)
        self.update_results()

    def update_results(self):
        i: int
        row: int
        cell: int
        action: Optional[Action]
        description: str
        index: int
        for i, row in enumerate(dpg.get_item_children(self.results_table, slot=1)):
            cell = dpg.get_item_children(row, slot=1)[0]
            action = None
            description = ""
            if self.result_index < 4:
                # A row in the upper half
                index = i
                dpg.bind_item_theme(
                    cell,
                    themes.command_palette.result_highlighted
                    if i == self.result_index
                    else themes.command_palette.result,
                )
            elif len(self.valid_actions) - self.result_index < 5:
                # A row in the lower half
                index = len(self.valid_actions) + i - self.num_rows
                dpg.bind_item_theme(
                    cell,
                    themes.command_palette.result_highlighted
                    if i
                    == self.num_rows - (len(self.valid_actions) - self.result_index)
                    else themes.command_palette.result,
                )
            else:
                # The row in the middle
                index = self.result_index + i - 4
                dpg.bind_item_theme(
                    cell,
                    themes.command_palette.result_highlighted
                    if i == 4
                    else themes.command_palette.result,
                )
            if index >= 0 and index < len(self.valid_actions):
                action, description = self.valid_actions[index]
            dpg.set_item_label(cell, description)
            dpg.set_item_user_data(cell, action)

    def navigate_result(self, index: Optional[int] = None, step: Optional[int] = None):
        assert type(index) is int or index is None, index
        assert type(step) is int or step is None, step
        if index is None and step is None:
            return
        elif index is not None:
            if index >= len(self.valid_actions):
                return
        elif step is not None:
            index = self.result_index + step
            if index < 0:
                index = 0
            elif index >= len(self.valid_actions):
                index = len(self.valid_actions) - 1
        assert index is not None
        self.result_index = index
        self.update_results()

    def perform_action(self, action: Optional[Action] = None, click: bool = False):
        if click and action is None:
            return
        self.hide()
        if not click:
            action = self.valid_actions[self.result_index][0]
        assert type(action) is Action, action
        if action in self.action_history:
            self.action_history.remove(action)
        self.action_history.insert(0, action)
        self.keybinding_handler.perform_action(
            action,
            self.current_contexts[-1],
            self.current_project,
            self.current_tab,
        )



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from types import SimpleNamespace


circuit_editor = SimpleNamespace(
    **{
        "cdc_input": """
Circuit description code (CDC) is a text-based representation of a circuit. The inputted value will be validated and then recreated below using nodes. Note that not all circuits can be defined using a CDC.

---- Basic syntax ----
Elements are represented with one or more letters (e.g. 'R' is a resistor). Elements defined using the basic syntax will have the default initial values for parameters and lower/upper limits for those parameters.

Two or more elements that are connected in series are placed one after another and enclosed with square brackets (e.g. '[RC]' is a resistor and a capacitor connected in series).

Two or more elements that are connected in parallel are placed one after another and enclosed with parentheses (e.g. '(RC)' is a resistor and a capacitor connected in parallel).

Series and parallel connections can be nested inside other series and parallel connections (e.g. "R([RW]C)" is the CDC for the Randles circuit).


---- Extended syntax ----
DearEIS supports an extended syntax for CDCs and thus values, lower/upper limits, and labels can also be defined for each element in a CDC.

A resistor with an initial value of 25 kiloohms can de defined as 'R{R=25e3}' or 'R{R=2.5e4}'. Note that numeric values must use periods/dots as decimal separators and no thousands separators are allowed at all.

The initial value can be defined as a fixed value by appending the numeric value with the lower- or upper-case letter 'f' (i.e. 'R{R=25e3f}' or 'R{R=25e3F}').

A lower limit can be defined by appending a forward slash and then a numeric value (e.g. 'R{R=25e3/20e3}' sets 20 kiloohms as the lower limit).

An upper limit can be defined in a similar way as a lower limit (e.g. 'R{R=25e3/20e3/30e3}' sets 30 kiloohms as the upper limit and 20 kiloohms as the lower limit). The lower limit can also be omitted (e.g. 'R{R=25e3//30e3}' sets 30 kiloohms as the upper limit while leaving the lower boundary unlimited).

Some elements have multiple parameters and in that case multiple parameter definitions can be included in the CDC by separating the parameter definitions with commas (e.g. 'Q{Y=1e-5//1e-4,n=0.8F}').

A label can be defined by appending a colon and then a string of text (e.g. 'R{R=25e3//30e3:example}'). The parameter values need not be defined at all (i.e. 'R{:another example}' is also valid).
    """.strip(),
        "element_combo": """
This drop-down list can be used to select an element that you want to add as a node to the window below once the 'Add' button is clicked.

The 'Add dummy' button adds a dummy node that can be used as a junction point, which may be necessary when representing certain circuits using nodes.
    """.strip(),
        "basic_cdc": """
This is the basic CDC output generated based on the nodes and connections in the window above.

This output may be accepted by another program that accepts a CDC as input. However, some programs may use different identifiers for some elements. Thus, this output may require adjustments for it to be accepted by some other programs.
    """.strip(),
        "extended_cdc": """
This is the extended CDC output generated based on the nodes and connections in the window above.

This output is unlikely to be accepted by another program that accepts a CDC as input.
    """.strip(),
        "status": """
Potential issues with the circuit, which is defined using nodes and connections in the window above, are presented here.
    """.strip(),
        "dummy_node": """
A dummy node that can be used as a junction when required to construct a circuit.
    """.strip(),
    }
)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)
import dearpygui.dearpygui as dpg
from deareis.signals import Signal
import deareis.signals as signals
from deareis.data import (
    DRTResult,
    DataSet,
    FitResult,
    PlotSettings,
    SimulationResult,
    TestResult,
)
from deareis.utility import calculate_window_position_dimensions
from deareis.tooltips import attach_tooltip
import deareis.tooltips as tooltips
from deareis.data import Project
import deareis.themes as themes
from deareis.keybindings import (
    is_alt_down,
    is_control_down,
)


class SeriesBefore:
    def __init__(
        self,
        series: Union[DataSet, TestResult, DRTResult, FitResult, SimulationResult],
        settings: PlotSettings,
        marker_lookup: Dict[int, str],
        toggle_callback: Callable,
    ):
        assert type(series) in [
            DRTResult,
            DataSet,
            FitResult,
            SimulationResult,
            TestResult,
        ], series
        assert type(settings) is PlotSettings, settings
        assert type(marker_lookup) is dict, marker_lookup
        with dpg.table_row(user_data=series.uuid):
            dpg.add_checkbox(
                default_value=True,
                callback=lambda s, a, u: toggle_callback(a, u),
                user_data=series.uuid,
            )
            attach_tooltip(tooltips.plotting.copy_appearance_toggle)
            label = settings.get_series_label(series.uuid) or series.get_label()
            dpg.add_text(label)
            attach_tooltip(
                label
                + ("" if label == series.get_label() else f"\n{series.get_label()}")
            )
            dpg.add_color_edit(
                default_value=settings.get_series_color(series.uuid),
                enabled=False,
                no_picker=True,
                alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                no_inputs=True,
            )
            dpg.add_text(
                marker_lookup.get(settings.get_series_marker(series.uuid), "None")
            )
            dpg.add_checkbox(
                default_value=settings.get_series_line(series.uuid),
                enabled=False,
            )


class SeriesAfter:
    def __init__(
        self,
        series,
        source,
        destination,
        copy_labels,
        copy_colors,
        copy_markers,
        copy_lines,
        marker_lookup,
        parent,
    ):
        with dpg.table_row(parent=parent, user_data=series.uuid):
            label: str = (
                source.get_series_label(series.uuid)
                if copy_labels
                else destination.get_series_label(series.uuid)
            )
            dpg.add_text(
                label or series.get_label(),
                user_data=label,
            )
            attach_tooltip(
                label
                or series.get_label()
                + (
                    ""
                    if (label or series.get_label()) == series.get_label()
                    else f"\n{series.get_label()}"
                )
            )
            color: List[float] = (
                source.get_series_color(series.uuid)
                if copy_colors
                else destination.get_series_color(series.uuid)
            )
            dpg.add_color_edit(
                default_value=color,
                enabled=False,
                no_picker=True,
                alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                no_inputs=True,
                user_data=color,
            )
            marker: int = (
                source.get_series_marker(series.uuid)
                if copy_markers
                else destination.get_series_marker(series.uuid)
            )
            dpg.add_text(
                marker_lookup.get(marker, "None"),
                user_data=marker,
            )
            show_line: bool = (
                source.get_series_line(series.uuid)
                if copy_lines
                else destination.get_series_line(series.uuid)
            )
            dpg.add_checkbox(
                default_value=show_line,
                enabled=False,
                user_data=show_line,
            )


class CopyPlotAppearance:
    def __init__(self, settings: PlotSettings, project: Project):
        assert type(settings) is PlotSettings, settings
        assert type(project) is Project, project
        self.settings: PlotSettings = settings
        self.marker_lookup: Dict[int, str] = {
            v: k for k, v in themes.PLOT_MARKERS.items()
        }
        self.plot_lookup: Dict[str, PlotSettings] = {
            _.get_label(): _ for _ in project.get_plots() if _ != settings
        }
        self.labels: List[str] = list(sorted(self.plot_lookup.keys()))
        self.series_checkboxes: Dict[str, bool] = {
            _: True for _ in settings.series_order
        }
        self.data_sets: List[DataSet] = project.get_data_sets()
        self.tests: Dict[str, List[TestResult]] = project.get_all_tests()
        self.drts: Dict[str, List[DRTResult]] = project.get_all_drts()
        self.fits: Dict[str, List[FitResult]] = project.get_all_fits()
        self.simulations: List[SimulationResult] = project.get_simulations()
        x: int
        y: int
        w: int
        h: int
        x, y, w, h = calculate_window_position_dimensions(720, 540)
        self.window: int = dpg.generate_uuid()
        with dpg.window(
            label="Copy appearance settings",
            modal=True,
            no_resize=True,
            pos=(
                x,
                y,
            ),
            width=w,
            height=h,
            on_close=self.close,
            tag=self.window,
        ):
            with dpg.group(horizontal=True):
                dpg.add_text("Source")
                self.source_combo: int = dpg.generate_uuid()
                dpg.add_combo(
                    items=self.labels,
                    default_value=self.labels[0],
                    width=-1,
                    callback=self.change_source,
                    tag=self.source_combo,
                )
            with dpg.group(horizontal=True):
                with dpg.group():
                    dpg.add_text("Before")
                    with dpg.child_window(width=348, height=-24):
                        with dpg.table(
                            borders_outerV=True,
                            borders_outerH=True,
                            borders_innerV=True,
                            borders_innerH=True,
                            scrollY=True,
                            freeze_rows=1,
                            width=-1,
                            height=18 + 23 * len(self.settings.series_order),
                        ):
                            dpg.add_table_column(label="", width_fixed=True)
                            dpg.add_table_column(label="Label")
                            dpg.add_table_column(label="Color", width_fixed=True)
                            dpg.add_table_column(label="Marker", width_fixed=True)
                            dpg.add_table_column(label="Line", width_fixed=True)
                            uuid: str
                            for uuid in settings.series_order:
                                series: Optional[
                                    Union[
                                        DataSet, TestResult, FitResult, SimulationResult
                                    ]
                                ]
                                series = settings.find_series(
                                    uuid,
                                    self.data_sets,
                                    self.tests,
                                    self.drts,
                                    self.fits,
                                    self.simulations,
                                )
                                assert series is not None
                                SeriesBefore(
                                    series,
                                    settings,
                                    self.marker_lookup,
                                    self.toggle_series,
                                )
                with dpg.group():
                    dpg.add_text("After")
                    with dpg.child_window(width=348, height=-24):
                        self.preview_table: int = dpg.generate_uuid()
                        with dpg.table(
                            borders_outerV=True,
                            borders_outerH=True,
                            borders_innerV=True,
                            borders_innerH=True,
                            scrollY=True,
                            freeze_rows=1,
                            width=-1,
                            height=18 + 23 * len(self.settings.series_order),
                            tag=self.preview_table,
                        ):
                            dpg.add_table_column(label="Label")
                            dpg.add_table_column(label="Color", width_fixed=True)
                            dpg.add_table_column(label="Marker", width_fixed=True)
                            dpg.add_table_column(label="Line", width_fixed=True)
            with dpg.group(horizontal=True):
                dpg.add_button(
                    label="Accept",
                    callback=self.accept,
                )
                dpg.add_spacer(width=354)
                self.labels_checkbox: int = dpg.generate_uuid()
                dpg.add_checkbox(
                    label="Labels",
                    default_value=True,
                    tag=self.labels_checkbox,
                    callback=lambda s, a, u: self.change_source(),
                )
                attach_tooltip(tooltips.plotting.copy_appearance_labels)
                self.colors_checkbox: int = dpg.generate_uuid()
                dpg.add_checkbox(
                    label="Colors",
                    default_value=True,
                    tag=self.colors_checkbox,
                    callback=lambda s, a, u: self.change_source(),
                )
                attach_tooltip(tooltips.plotting.copy_appearance_colors)
                self.markers_checkbox: int = dpg.generate_uuid()
                dpg.add_checkbox(
                    label="Markers",
                    default_value=True,
                    tag=self.markers_checkbox,
                    callback=lambda s, a, u: self.change_source(),
                )
                attach_tooltip(tooltips.plotting.copy_appearance_markers)
                self.lines_checkbox: int = dpg.generate_uuid()
                dpg.add_checkbox(
                    label="Lines",
                    default_value=True,
                    tag=self.lines_checkbox,
                    callback=lambda s, a, u: self.change_source(),
                )
                attach_tooltip(tooltips.plotting.copy_appearance_lines)
        self.key_handler: int = dpg.generate_uuid()
        with dpg.handler_registry(tag=self.key_handler):
            dpg.add_key_release_handler(
                key=dpg.mvKey_Escape,
                callback=self.close,
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Return,
                callback=lambda: self.accept(keybinding=True),
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Prior,
                callback=lambda: self.cycle_source(-1),
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Next,
                callback=lambda: self.cycle_source(1),
            )
        signals.emit(Signal.BLOCK_KEYBINDINGS, window=self.window, window_object=self)
        self.change_source()

    def close(self):
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)
        dpg.delete_item(self.window)
        dpg.delete_item(self.key_handler)

    def accept(self, keybinding: bool = False):
        if keybinding is True and not (
            is_control_down()
            if dpg.get_platform() == dpg.mvPlatform_Windows
            else is_alt_down()
        ):
            return
        dpg.hide_item(self.window)
        changes: Dict[str, Tuple[str, List[float], int, bool]] = {}
        row: int
        for row in dpg.get_item_children(self.preview_table, slot=1):
            uuid: str = dpg.get_item_user_data(row)
            label: str = ""
            color: List[float] = [255.0, 255.0, 255.0, 255.0]
            marker: int = -1
            line: bool = False
            i: int
            item: int
            for i, item in enumerate(dpg.get_item_children(row, slot=1)):
                item_type: str = dpg.get_item_type(item)
                if i == 0:
                    assert "::mvText" in item_type, item_type
                    label = dpg.get_item_user_data(item)
                elif i == 1:
                    assert "::mvTooltip" in item_type, item_type
                elif i == 2:
                    assert "::mvColorEdit" in item_type, item_type
                    color = dpg.get_item_user_data(item)
                elif i == 3:
                    assert "::mvText" in item_type, item_type
                    marker = dpg.get_item_user_data(item)
                elif i == 4:
                    assert "::mvCheckbox" in item_type, item_type
                    line = dpg.get_item_user_data(item)
            changes[uuid] = (
                label,
                color,
                marker,
                line,
            )
        signals.emit(
            Signal.COPY_PLOT_APPEARANCE_SETTINGS,
            changes=changes,
            settings=self.settings,
        )
        self.close()

    def cycle_source(self, step: int):
        assert type(step) is int, step
        index: int = self.labels.index(dpg.get_value(self.source_combo)) + step
        dpg.set_value(self.source_combo, self.labels[index % len(self.labels)])
        self.change_source()

    def change_source(self):
        label: str = dpg.get_value(self.source_combo)
        source: PlotSettings = self.plot_lookup[label]
        dpg.delete_item(self.preview_table, children_only=True, slot=1)
        copy_labels: bool = dpg.get_value(self.labels_checkbox)
        copy_colors: bool = dpg.get_value(self.colors_checkbox)
        copy_markers: bool = dpg.get_value(self.markers_checkbox)
        copy_lines: bool = dpg.get_value(self.lines_checkbox)
        for uuid in self.settings.series_order:
            SeriesAfter(
                self.settings.find_series(
                    uuid,
                    self.data_sets,
                    self.tests,
                    self.drts,
                    self.fits,
                    self.simulations,
                ),
                source
                if uuid in source.themes and self.series_checkboxes[uuid]
                else self.settings,
                self.settings,
                copy_labels,
                copy_colors,
                copy_markers,
                copy_lines,
                self.marker_lookup,
                self.preview_table,
            )

    def toggle_series(self, state: bool, uuid: str):
        self.series_checkboxes[uuid] = state
        self.change_source()



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from threading import Timer
from traceback import format_exc
from typing import Any, List, Optional
import dearpygui.dearpygui as dpg
from deareis.tooltips import attach_tooltip
from deareis.utility import calculate_window_position_dimensions
from deareis.enums import Action, action_descriptions
from deareis.signals import Signal
import deareis.signals as signals
from deareis.keybindings import Keybinding, dpg_to_string
from deareis.config import DEFAULT_KEYBINDINGS


class KeybindingTable:
    def __init__(self, key_filter_input: int, description_filter_input: int, state):
        self.key_filter_input: int = key_filter_input
        self.description_filter_input: int = description_filter_input
        self.state = state
        self.remapping: bool = False
        self.info_window: int = dpg.generate_uuid()
        with dpg.child_window(width=-1, height=-26, tag=self.info_window):
            self.info_text: int = dpg.generate_uuid()
            dpg.add_text(
                "",
                tag=self.info_text,
                user_data="Press a key or Esc to clear the current key mapping...",
                wrap=680,
            )
        self.table: int = dpg.generate_uuid()
        with dpg.table(
            borders_outerV=True,
            borders_outerH=True,
            borders_innerV=True,
            borders_innerH=True,
            scrollY=True,
            height=18 + 23 * 19,
            freeze_rows=1,
            tag=self.table,
        ):
            dpg.add_table_column(
                label=" A",
                width_fixed=True,
            )
            attach_tooltip("Whether or not the Alt modifier key is required.")
            dpg.add_table_column(
                label=" C",
                width_fixed=True,
            )
            attach_tooltip("Whether or not the Ctrl modifier key is required.")
            dpg.add_table_column(
                label=" S",
                width_fixed=True,
            )
            attach_tooltip("Whether or not the Shift modifier key is required.")
            self.key_padding: int = max(list(map(len, list(dpg_to_string.values()))))
            dpg.add_table_column(
                label="Key".ljust(self.key_padding),
                width_fixed=True,
            )
            attach_tooltip(
                "Press a button in this column to remap the key. Note that not all keys are supported. Press 'Esc' while remapping to clear the key instead of remapping it to another key."
            )
            dpg.add_table_column(
                label="Description",
            )
        self.reset()

    def is_remapping(self) -> bool:
        return self.remapping

    def populate(self):
        dpg.delete_item(self.table, children_only=True, slot=1)
        action: Action
        description: str
        for action, description in action_descriptions.items():
            kb: Optional[Keybinding] = self.find_keybinding(action)
            filter_key: str = "|".join(
                [str(kb) if kb else "", description.lower()]
            ).lower()
            with dpg.table_row(
                filter_key=filter_key,
                user_data=action,
                parent=self.table,
            ):
                dpg.add_checkbox(
                    default_value=kb.mod_alt if kb else False,
                    callback=lambda s, a, u: self.toggle_mod_key(s, a, u["kb"]),
                    user_data={
                        "kb": kb,
                        "mod": "alt",
                    },
                )
                dpg.add_checkbox(
                    default_value=kb.mod_ctrl if kb else False,
                    callback=lambda s, a, u: self.toggle_mod_key(s, a, u["kb"]),
                    user_data={
                        "kb": kb,
                        "mod": "ctrl",
                    },
                )
                dpg.add_checkbox(
                    default_value=kb.mod_shift if kb else False,
                    callback=lambda s, a, u: self.toggle_mod_key(s, a, u["kb"]),
                    user_data={
                        "kb": kb,
                        "mod": "shift",
                    },
                )
                dpg.add_button(
                    label=(dpg_to_string.get(kb.key, "") if kb else "").ljust(
                        self.key_padding
                    ),
                    callback=lambda s, a, u: self.begin_remap(**u),
                    user_data={
                        "kb": kb,
                        "action": action,
                    },
                )
                short_description: str = description.split("\n")[0]
                description_limit: int = 64
                if len(short_description) > description_limit:
                    short_description = short_description[:description_limit] + "..."
                if short_description.endswith(":"):
                    short_description = short_description[:-1] + "..."
                dpg.add_text(short_description)
                attach_tooltip(description)
        self.filter()

    def filter(self):
        key: str = dpg.get_value(self.key_filter_input).strip().lower()
        description: str = dpg.get_value(self.description_filter_input).strip().lower()
        row: int
        for row in dpg.get_item_children(self.table, slot=1):
            key_filter: str
            description_filter: str
            key_filter, description_filter = dpg.get_item_filter_key(row).split("|")
            if key and description:
                if key in key_filter and description in description_filter:
                    dpg.show_item(row)
                else:
                    dpg.hide_item(row)
            elif key:
                if key in key_filter:
                    dpg.show_item(row)
                else:
                    dpg.hide_item(row)
            elif description:
                if description in description_filter:
                    dpg.show_item(row)
                else:
                    dpg.hide_item(row)
            else:
                dpg.show_item(row)

    def find_keybinding(self, action: Action) -> Optional[Keybinding]:
        kb: Keybinding
        for kb in self.state.config.keybindings:
            if kb.action == action:
                return kb
        return None

    def toggle_mod_key(self, sender: int, state: bool, kb: Optional[Keybinding]):
        if kb is None:
            return
        keybindings: List[Keybinding] = self.state.config.keybindings.copy()
        keybindings.remove(kb)
        keybindings.append(
            Keybinding(
                kb.key,
                self.has_mod_alt(kb.action),
                self.has_mod_ctrl(kb.action),
                self.has_mod_shift(kb.action),
                kb.action,
            )
        )
        try:
            self.state.config.validate_keybindings(keybindings)
        except Exception as e:
            dpg.hide_item(self.table)
            dpg.set_value(self.info_text, str(e))
            dpg.show_item(self.info_window)
            # dpg.set_value(self.info_text, format_exc())
            t: Timer = Timer(5, self.reset)
            t.start()
            return
        try:
            self.state.config.keybindings = keybindings
            self.state.keybinding_handler.register(keybindings)
            self.reset()
        except Exception:
            signals.emit(Signal.SHOW_ERROR_MESSAGE, traceback=format_exc())

    def has_mod_alt(self, action: Action) -> bool:
        return self.has_mod(action, "alt")

    def has_mod_ctrl(self, action: Action) -> bool:
        return self.has_mod(action, "ctrl")

    def has_mod_shift(self, action: Action) -> bool:
        return self.has_mod(action, "shift")

    def has_mod(self, action: Action, mod: str) -> bool:
        row: int
        for row in dpg.get_item_children(self.table, slot=1):
            if dpg.get_item_user_data(row) != action:
                continue
            cell: int
            for cell in dpg.get_item_children(row, slot=1):
                user_data: Any = dpg.get_item_user_data(cell)
                if type(user_data) is not dict:
                    continue
                elif user_data.get("mod") == mod:
                    return dpg.get_value(cell)
        return False

    def begin_remap(self, kb: Optional[Keybinding], action: Action):
        self.remapping = True
        self.key_handler: int = dpg.generate_uuid()
        with dpg.handler_registry(tag=self.key_handler):
            key: int
            for key in dpg_to_string:
                dpg.add_key_release_handler(
                    key=key,
                    callback=lambda s, a, u: self.end_remap(a, kb, action),
                )
        dpg.hide_item(self.table)
        dpg.show_item(self.info_window)

    def end_remap(self, key: int, kb: Optional[Keybinding], action: Action):
        if dpg.does_item_exist(self.key_handler):
            dpg.delete_item(self.key_handler)
        self.remapping = False
        keybindings: List[Keybinding] = self.state.config.keybindings.copy()
        if kb is not None:
            keybindings.remove(kb)
        if key != dpg.mvKey_Escape:
            kb = Keybinding(
                key,
                self.has_mod_alt(action),
                self.has_mod_ctrl(action),
                self.has_mod_shift(action),
                action,
            )
            keybindings.append(kb)  # type: ignore
        try:
            self.state.config.validate_keybindings(keybindings)
        except Exception as e:
            dpg.set_value(self.info_text, str(e))
            # dpg.set_value(self.info_text, format_exc())
            t: Timer = Timer(5, self.reset)
            t.start()
            return
        try:
            self.state.config.keybindings = keybindings
            self.state.keybinding_handler.register(keybindings)
            self.reset()
        except Exception:
            signals.emit(Signal.SHOW_ERROR_MESSAGE, traceback=format_exc())

    def reset(self):
        dpg.set_value(self.info_text, dpg.get_item_user_data(self.info_text))
        dpg.hide_item(self.info_window)
        dpg.show_item(self.table)
        self.populate()


class KeybindingRemapping:
    def __init__(self, state):
        self.state = state
        x: int
        y: int
        w: int
        h: int
        x, y, w, h = calculate_window_position_dimensions(720, 540)
        self.window: int = dpg.generate_uuid()
        with dpg.window(
            label="Settings - keybindings",
            modal=True,
            pos=(
                x,
                y,
            ),
            width=w,
            height=h,
            no_move=False,
            no_resize=True,
            on_close=self.close_window,
            tag=self.window,
        ):
            key_filter_input: int = dpg.generate_uuid()
            description_filter_input: int = dpg.generate_uuid()
            with dpg.group(horizontal=True):
                dpg.add_input_text(
                    callback=lambda s, a, u: self.table.filter(),
                    hint="Filter key(s)",
                    width=200,
                    tag=key_filter_input,
                )
                attach_tooltip(
                    """
Filter based on keys or modifiers ('alt', 'ctrl', or 'shift'). Multiple modifiers and a key can be filtered by using '+' as a separator (e.g. 'alt+shift+n').
                """.strip()
                )
                dpg.add_input_text(
                    callback=lambda s, a, u: self.table.filter(),
                    hint="Filter description(s)",
                    width=-1,
                    tag=description_filter_input,
                )
                attach_tooltip(
                    """
Filter based on descriptions.
                """.strip()
                )
            self.table: KeybindingTable = KeybindingTable(
                key_filter_input, description_filter_input, state
            )
            with dpg.group(horizontal=True):
                dpg.add_button(label="Clear all", callback=self.clear_all)
                dpg.add_button(label="Reset", callback=self.reset)
        self.key_handler: int = dpg.generate_uuid()
        with dpg.handler_registry(tag=self.key_handler):
            dpg.add_key_release_handler(
                key=dpg.mvKey_Escape,
                callback=self.close_window,
            )
        signals.emit(Signal.BLOCK_KEYBINDINGS, window=self.window, window_object=self)

    def clear_all(self):
        self.state.config.keybindings = []
        self.state.keybinding_handler.register(self.state.config.keybindings)
        self.table.populate()

    def reset(self):
        self.state.config.keybindings = DEFAULT_KEYBINDINGS.copy()
        self.state.keybinding_handler.register(self.state.config.keybindings)
        self.table.populate()

    def close_window(self):
        if self.table.is_remapping():
            return
        if dpg.does_item_exist(self.window):
            dpg.delete_item(self.window)
        if dpg.does_item_exist(self.key_handler):
            dpg.delete_item(self.key_handler)
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    Callable,
    List,
    Optional,
)
from numpy import (
    allclose,
    ndarray,
)
import dearpygui.dearpygui as dpg
from deareis.gui.plots import Nyquist
from deareis.tooltips import attach_tooltip
from deareis.themes import (
    PLOT_MARKERS,
    VIBRANT_COLORS,
    create_plot_series_theme,
)
import deareis.themes as themes
from deareis.utility import calculate_window_position_dimensions
from deareis.signals import Signal
import deareis.signals as signals
from deareis.data import DataSet
from deareis.keybindings import (
    is_alt_down,
    is_control_down,
)


class AverageDataSets:
    def __init__(self, data_sets: List[DataSet], callback: Callable):
        assert type(data_sets) is list and all(
            map(lambda _: type(_) is DataSet, data_sets)
        )
        self.data_sets: List[DataSet] = data_sets
        self.callback: Callable = callback
        self.plot_themes: List[int] = list(
            map(
                self.get_plot_theme,
                range(0, 12),
            )
        )
        self.final_data: Optional[DataSet] = None
        x: int
        y: int
        w: int
        h: int
        x, y, w, h = calculate_window_position_dimensions()
        self.window: int = dpg.generate_uuid()
        with dpg.window(
            label="Average of multiple data sets",
            modal=True,
            pos=(
                x,
                y,
            ),
            width=w,
            height=h,
            tag=self.window,
            on_close=self.close,
        ):
            with dpg.group(horizontal=True):
                with dpg.group():
                    with dpg.child_window(border=False, width=300, height=-24):
                        with dpg.group(horizontal=True):
                            dpg.add_text("Label")
                            self.label_input: int = dpg.generate_uuid()
                            self.final_data_series: int = -1
                            dpg.add_input_text(
                                hint="REQUIRED",
                                default_value="Average",
                                width=-1,
                                tag=self.label_input,
                                callback=lambda s, a, u: self.update_label(a),
                            )
                        self.dataset_table: int = dpg.generate_uuid()
                        with dpg.table(
                            borders_outerV=True,
                            borders_outerH=True,
                            borders_innerV=True,
                            borders_innerH=True,
                            scrollY=True,
                            freeze_rows=1,
                            width=-1,
                            height=-1,
                            tag=self.dataset_table,
                        ):
                            dpg.add_table_column(label="", width_fixed=True)
                            dpg.add_table_column(label="Label", width_fixed=True)
                            data: DataSet
                            for data in self.data_sets:
                                with dpg.table_row():
                                    dpg.add_checkbox(
                                        callback=lambda: self.update_preview([])
                                    )
                                    label: str = data.get_label()
                                    dpg.add_text(label)
                                    attach_tooltip(label)
                    dpg.add_button(label="Accept", callback=self.accept)
                self.nyquist_plot: Nyquist = Nyquist(
                    width=-1,
                    height=-12,
                    legend_horizontal=False,
                    legend_outside=False,
                    legend_location=dpg.mvPlot_Location_NorthEast,
                )
        self.key_handler: int = dpg.generate_uuid()
        with dpg.handler_registry(tag=self.key_handler):
            dpg.add_key_release_handler(
                key=dpg.mvKey_Escape,
                callback=self.close,
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Return,
                callback=lambda: self.accept(keybinding=True),
            )
        self.update_preview([])

    def close(self):
        dpg.hide_item(self.window)
        dpg.delete_item(self.window)
        dpg.delete_item(self.key_handler)
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)

    def accept(self, keybinding: bool = False):
        if self.final_data is None:
            return
        if keybinding is True and not (
            is_control_down()
            if dpg.get_platform() == dpg.mvPlatform_Windows
            else is_alt_down()
        ):
            return
        label: str = dpg.get_value(self.label_input).strip()
        if label == "":
            return
        self.final_data.set_label(label)
        self.callback(self.final_data)
        self.close()

    def get_selection(self) -> List[DataSet]:
        indices: List[int] = []
        data_sets: List[DataSet] = []
        i: int
        row: int
        for i, row in enumerate(dpg.get_item_children(self.dataset_table, slot=1)):
            column: int
            for column in dpg.get_item_children(row, slot=1):
                assert dpg.get_item_type(column).endswith("Checkbox")
                if dpg.get_value(column) is True:
                    data_sets.append(self.data_sets[i])
                    indices.append(i)
                break
        if data_sets:
            frequency: ndarray = data_sets[0].get_frequency(masked=None)
            for i, row in enumerate(dpg.get_item_children(self.dataset_table, slot=1)):
                if i in indices:
                    continue
                data: DataSet = self.data_sets[i]
                if not (
                    data.get_num_points(masked=None) == frequency.size
                    and allclose(data.get_frequency(masked=None), frequency)
                ):
                    dpg.hide_item(dpg.get_item_children(row, slot=1)[0])
        else:
            for i, row in enumerate(dpg.get_item_children(self.dataset_table, slot=1)):
                dpg.show_item(dpg.get_item_children(row, slot=1)[0])
        return data_sets

    def get_plot_theme(self, index: int) -> int:
        assert type(index) is int, index
        return create_plot_series_theme(
            VIBRANT_COLORS[index % len(VIBRANT_COLORS)],
            list(PLOT_MARKERS.values())[index % len(PLOT_MARKERS)],
        )

    def update_preview(self, data_sets: List[DataSet]):
        assert type(data_sets) is list and all(
            map(lambda _: type(_) is DataSet, data_sets)
        ), data_sets
        from_empty: bool = False
        if len(self.nyquist_plot.get_series()) == 0:
            from_empty = True
        self.nyquist_plot.clear()
        selection: List[DataSet] = self.get_selection()
        i: int
        data: DataSet
        real: ndarray
        imag: ndarray
        for i, data in enumerate(selection):
            real, imag = data.get_nyquist_data(masked=None)
            self.nyquist_plot.plot(
                real=real,
                imaginary=imag,
                label=data.get_label(),
                line=False,
                theme=self.plot_themes[i % 12],
            )
        self.final_data_series = -1
        if len(selection) > 1:
            try:
                self.final_data = DataSet.average(
                    selection,
                    label=dpg.get_value(self.label_input),
                )
                assert self.final_data is not None
                real, imag = self.final_data.get_nyquist_data(masked=None)
                self.final_data_series = self.nyquist_plot.plot(
                    real=real,
                    imaginary=imag,
                    label=self.final_data.get_label(),
                    line=True,
                    theme=themes.nyquist.data,
                )
            except AssertionError:
                self.final_data = None
        if from_empty:
            self.nyquist_plot.queue_limits_adjustment()

    def update_label(self, label: str):
        assert type(label) is str, label
        if self.final_data_series < 0 or not dpg.does_item_exist(
            self.final_data_series
        ):
            return
        # TODO: Apply a theme that attracts attention when no label is provided
        if label == "":
            pass
        else:
            pass
        dpg.set_item_label(self.final_data_series, label)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from dataclasses import dataclass
from typing import (
    Callable,
    Dict,
    List,
    Tuple,
)
from numpy import (
    angle,
    integer,
    issubdtype,
    ndarray,
)
from pandas import DataFrame
from pyimpspec import (
    Circuit,
    Element,
)
import pyimpspec
from pyimpspec.analysis.fitting import _interpolate
from deareis.utility import format_timestamp


VERSION: int = 2


def _parse_settings_v2(dictionary: dict) -> dict:
    assert type(dictionary) is dict
    return {
        "cdc": dictionary["cdc"],
        "min_frequency": dictionary["min_frequency"],
        "max_frequency": dictionary["max_frequency"],
        "num_per_decade": dictionary["num_per_decade"],
    }


def _parse_settings_v1(dictionary: dict) -> dict:
    assert type(dictionary) is dict
    return {
        "cdc": dictionary["cdc"],
        "min_frequency": dictionary["min_frequency"],
        "max_frequency": dictionary["max_frequency"],
        "num_per_decade": dictionary["num_freq_per_dec"],
    }


@dataclass(frozen=True)
class SimulationSettings:
    """
    A class to store the settings used to perform a simulation.

    Parameters
    ----------
    cdc: str
        The circuit description code (CDC) for the circuit to simulate.

    min_frequency: float
        The minimum frequency (in hertz) to simulate.

    max_frequency: float
        The maximum frequency (in hertz) to simulate.

    num_per_decade: int
        The number of frequencies per decade to simulate.
        The frequencies are distributed logarithmically within the inclusive boundaries defined by `min_frequency` and `max_frequency`.
    """

    cdc: str
    min_frequency: float
    max_frequency: float
    num_per_decade: int

    def __repr__(self) -> str:
        return f"SimulationSettings ({hex(id(self))})"

    @classmethod
    def from_dict(Class, dictionary: dict) -> "SimulationSettings":
        """
        Create an instance from a dictionary.
        """
        assert type(dictionary) is dict
        assert "version" in dictionary
        version: int = dictionary["version"]
        assert version <= VERSION, f"{version=} > {VERSION=}"
        parsers: Dict[int, Callable] = {
            1: _parse_settings_v1,
            2: _parse_settings_v2,
        }
        assert version in parsers, f"{version=} not in {parsers.keys()=}"
        return Class(**parsers[version](dictionary))

    def to_dict(self) -> dict:
        """
        Return a dictionary that can be used to recreate an instance.
        """
        return {
            "version": VERSION,
            "cdc": self.cdc,
            "min_frequency": self.min_frequency,
            "max_frequency": self.max_frequency,
            "num_per_decade": self.num_per_decade,
        }


def _parse_result_v2(dictionary: dict) -> dict:
    assert type(dictionary) is dict
    return {
        "uuid": dictionary["uuid"],
        "timestamp": dictionary["timestamp"],
        "circuit": pyimpspec.parse_cdc(dictionary["circuit"]),
        "settings": SimulationSettings.from_dict(dictionary["settings"]),
    }


def _parse_result_v1(dictionary: dict) -> dict:
    assert type(dictionary) is dict
    return {
        "uuid": dictionary["uuid"],
        "timestamp": dictionary["timestamp"],
        "circuit": pyimpspec.parse_cdc(dictionary["circuit"]),
        "settings": SimulationSettings.from_dict(dictionary["settings"]),
    }


@dataclass
class SimulationResult:
    """
    A class containing the result of a simulation.

    Parameters
    ----------
    uuid: str
        The universally unique identifier assigned to this result.

    timestamp: float
        The Unix time (in seconds) for when the simulation was performed.

    circuit: Circuit
        The simulated circuit.

    settings: SimulationSettings
        The settings that were used to perform the simulation.
    """

    uuid: str
    timestamp: float
    circuit: Circuit
    settings: SimulationSettings

    def __post_init__(self):
        self._cached_frequency: Dict[int, ndarray] = {}
        self._cached_impedance: Dict[int, ndarray] = {}
        self._frequency: ndarray = self.get_frequency(
            num_per_decade=self.settings.num_per_decade
        )
        self._impedance: ndarray = self.get_impedance(
            num_per_decade=self.settings.num_per_decade
        )

    def __repr__(self) -> str:
        return f"SimulationResult ({self.get_label()}, {hex(id(self))})"

    @classmethod
    def from_dict(Class, dictionary: dict) -> "SimulationResult":
        """
        Create an instance from a dictionary.
        """
        assert type(dictionary) is dict
        assert "version" in dictionary
        version: int = dictionary["version"]
        assert version <= VERSION, f"{version=} > {VERSION=}"
        parsers: Dict[int, Callable] = {
            1: _parse_result_v1,
            2: _parse_result_v2,
        }
        assert version in parsers, f"{version=} not in {parsers.keys()=}"
        return Class(**parsers[version](dictionary))

    def to_dict(self) -> dict:
        """
        Return a dictionary that can be used to recreate an instance.
        """
        return {
            "version": VERSION,
            "uuid": self.uuid,
            "timestamp": self.timestamp,
            "circuit": self.circuit.to_string(12),
            "settings": self.settings.to_dict(),
        }

    def to_dataframe(self) -> DataFrame:
        """
        Get a `pandas.DataFrame` instance containing a table of element parameters.
        """
        element_labels: List[str] = []
        parameter_labels: List[str] = []
        values: List[float] = []
        element: Element
        for element in sorted(
            self.circuit.get_elements(flattened=True),
            key=lambda _: _.get_identifier(),
        ):
            parameters: Dict[str, float] = element.get_parameters()
            parameter_label: str
            for parameter_label in sorted(parameters.keys()):
                element_labels.append(element.get_label())
                parameter_labels.append(parameter_label)
                values.append(parameters[parameter_label])
        return DataFrame.from_dict(
            {
                "Element": element_labels,
                "Parameter": parameter_labels,
                "Value": values,
            }
        )

    def get_label(self) -> str:
        """
        Generate a label for the result.
        """
        cdc: str = self.settings.cdc
        while "{" in cdc:
            i: int = cdc.find("{")
            j: int = cdc.find("}")
            cdc = cdc.replace(cdc[i : j + 1], "")
        if cdc.startswith("[") and cdc.endswith("]"):
            cdc = cdc[1:-1]
        timestamp: str = format_timestamp(self.timestamp)
        return f"{cdc} ({timestamp})"

    def get_frequency(self, num_per_decade: int = -1) -> ndarray:
        """
        Get an array of frequencies within the range of simulated frequencies.

        Parameters
        ----------
        num_per_decade: int = -1
            If the value is greater than zero, then logarithmically distributed frequencies will be generated within the range of frequencies defined by the minimum and maximum frequencies used to generate the original simulation result.
        """
        assert issubdtype(type(num_per_decade), integer), num_per_decade
        if num_per_decade > 0:
            if num_per_decade not in self._cached_frequency:
                self._cached_frequency.clear()
                self._cached_frequency[num_per_decade] = _interpolate(
                    [self.settings.min_frequency, self.settings.max_frequency],
                    num_per_decade,
                )
            return self._cached_frequency[num_per_decade]
        return self._frequency

    def get_impedance(self, num_per_decade: int = -1) -> ndarray:
        """
        Get the complex impedances produced by the simulated circuit within the range of frequencies used to generate the original simulation result.

        Parameters
        ----------
        num_per_decade: int = -1
            If the value is greater than zero, then logarithmically distributed frequencies will be generated within the range of simulated frequencies and used to calculate the impedance produced by the simulated circuit.
        """
        assert issubdtype(type(num_per_decade), integer), num_per_decade
        if num_per_decade > 0:
            if num_per_decade not in self._cached_impedance:
                self._cached_impedance.clear()
                self._cached_impedance[num_per_decade] = self.circuit.impedances(
                    self.get_frequency(num_per_decade)
                )
            return self._cached_impedance[num_per_decade]
        return self._impedance

    def get_nyquist_data(self, num_per_decade: int = -1) -> Tuple[ndarray, ndarray]:
        """
        Get the data required to plot the results as a Nyquist plot (-Z\" vs Z').

        Parameters
        ----------
        num_per_decade: int = -1
            If the value is greater than zero, then logarithmically distributed frequencies will be generated within the range of frequencies and used to calculate the impedance produced by the simulated circuit.
        """
        assert issubdtype(type(num_per_decade), integer), num_per_decade
        Z: ndarray = self.get_impedance(num_per_decade)
        return (
            Z.real,
            -Z.imag,
        )

    def get_bode_data(
        self, num_per_decade: int = -1
    ) -> Tuple[ndarray, ndarray, ndarray]:
        """
        Get the data required to plot the results as a Bode plot (|Z| and phi vs f).

        Parameters
        ----------
        num_per_decade: int = -1
            If the value is greater than zero, then logarithmically distributed frequencies will be generated within the range of frequencies and used to calculate the impedance produced by the fitted circuit.
        """
        assert issubdtype(type(num_per_decade), integer), num_per_decade
        f: ndarray = self.get_frequency(num_per_decade)
        Z: ndarray = self.get_impedance(num_per_decade)
        return (
            f,
            abs(Z),
            -angle(Z, deg=True),
        )
